<?php

namespace FiloBlu\Refilo\Remote;

use Exception;
use FiloBlu\Refilo\Remote\Entity\EntityProviderInterfaceFactory;
use FiloBlu\Refilo\Remote\Helper\TemplateString;
use FiloBlu\Refilo\Remote\Helper\TemplateStringFactory;
use FiloBlu\Refilo\Remote\IndexerConfiguration\IndexerConfigurationInterface;
use FiloBlu\Refilo\Remote\IndexerConfiguration\IndexerConfigurationInterfaceFactory;
use FiloBlu\Refilo\Remote\IndexerConfiguration\IndexerConfigurationRepositoryInterface;
use FiloBlu\Refilo\Remote\IndexerConfiguration\IndexerResolver;
use Magento\Framework\App\Config\ScopeConfigInterface;
use Magento\Framework\DataObject;
use Magento\Framework\DataObjectFactory;
use Magento\Framework\Exception\NoSuchEntityException;
use Magento\Store\Api\Data\StoreInterface;
use Magento\Store\Model\StoreManagerInterface;
use Psr\Log\LoggerInterface;
use RuntimeException;

/**
 * Class AbstractIndexer
 * @package FiloBlu\Refilo\Remote
 */
abstract class AbstractIndexer implements IndexerInterface
{
    /**
     * @var EntityProviderInterfaceFactory
     */
    protected $entityProviderFactory;
    /**
     * @var ScopeConfigInterface
     */
    protected $scopeConfig;
    /**
     * @var LoggerInterface
     */
    protected $logger;
    /**
     * @var DataObjectFactory
     */
    protected $dataObjectFactory;
    /**
     * @var StoreManagerInterface
     */
    private $storeManager;
    /**
     * @var string
     */
    private $name;
    /**
     * @var IndexerConfigurationRepositoryInterface
     */
    private $indexerConfigurationRepository;
    /**
     * @var IndexerConfigurationInterface
     */
    private $indexerConfiguration;
    /**
     * @var TemplateStringFactory
     */
    private $templateStringFactory;
    /**
     * @var  TemplateString
     */
    private $remoteIndexTemplate;
    /**
     * @var IndexerConfiguration\IndexerResolver
     */
    private $indexerResolver;
    /**
     * @var IndexerConfiguration\IndexerConfigurationInterfaceFactory
     */
    private $indexerConfigurationFactory;

    /**
     * AbstractIndexer constructor.
     * @param StoreManagerInterface $storeManager
     * @param EntityProviderInterfaceFactory $entityProviderFactory
     * @param IndexerConfigurationRepositoryInterface $indexerConfigurationRepository
     * @param TemplateStringFactory $templateStringFactory
     * @param IndexerResolver $indexerResolver
     * @param ScopeConfigInterface $scopeConfig
     * @param LoggerInterface $logger
     * @param DataObjectFactory $dataObjectFactory
     * @param IndexerConfigurationInterfaceFactory $indexerConfigurationFactory
     */
    public function __construct(
        StoreManagerInterface $storeManager,
        EntityProviderInterfaceFactory $entityProviderFactory,
        IndexerConfigurationRepositoryInterface $indexerConfigurationRepository,
        TemplateStringFactory $templateStringFactory,
        IndexerResolver $indexerResolver,
        ScopeConfigInterface $scopeConfig,
        LoggerInterface $logger,
        DataObjectFactory $dataObjectFactory,
        IndexerConfigurationInterfaceFactory $indexerConfigurationFactory
    )
    {
        $this->entityProviderFactory = $entityProviderFactory;
        $this->indexerConfigurationRepository = $indexerConfigurationRepository;
        $this->templateStringFactory = $templateStringFactory;
        $this->scopeConfig = $scopeConfig;
        $this->logger = $logger;
        $this->indexerResolver = $indexerResolver;
        $this->storeManager = $storeManager;
        $this->dataObjectFactory = $dataObjectFactory;
        $this->indexerConfigurationFactory = $indexerConfigurationFactory;
    }

    /**
     *
     */
    public function executeFull()
    {
        $this->executeList([]);
    }

    /**
     * @param array $ids
     */
    public function executeList(array $ids)
    {
        if(!$this->getIndexerConfiguration()->getEnabled()) {
            return;
        }

        try {
            $ids = array_unique($ids);
            $this->getEntitiesToRemove($ids);
            $this->beforeExecute($ids);
            $this->executeAction($ids);
            $this->postExecute($ids);
        } catch (Exception $exception) {
            $this->logger->error($exception);
        }
    }

    /**
     * @param array $ids
     * @return mixed
     */
    abstract public function getEntitiesToRemove(array $ids = []);

    /**
     * @param array $ids
     */
    abstract public function beforeExecute(array $ids = []);

    /**
     * @param array $ids
     * @return mixed
     */
    abstract public function executeAction(array $ids);

    /**
     * @param array $ids
     */
    abstract public function postExecute(array $ids = []);

    /**
     * @return StoreManagerInterface
     */
    public function getStoreManager()
    {
        return $this->storeManager;
    }

    /**
     * @param int $id
     * @throws NoSuchEntityException
     */
    public function executeRow($id)
    {
        $this->executeList([$id]);
    }

    /**
     * Invoked by indexer_update_all_views
     *
     * @param int[] $ids
     */
    public function execute($ids)
    {
        $this->executeList($ids);
    }

    /**
     * @return TemplateString
     * @throws NoSuchEntityException
     */
    public function getGetRemoteIndexTemplate(): TemplateString
    {
        if ($this->remoteIndexTemplate) {
            return $this->remoteIndexTemplate;
        }

        $indexName = $this->getIndexerConfiguration()->getIndexName();
        if (empty($indexName)) {
            throw new RuntimeException("Misconfigured Beehive indexer {$this->getName()}");
        }

        $projectId = strtolower($this->scopeConfig->getValue('filoblu/general/project_id'));

        /** @var TemplateString $indexTemplate */
        $this->remoteIndexTemplate = $this->templateStringFactory->create();

        $this->remoteIndexTemplate
            ->withString($indexName)
            ->withValue('%p', $projectId);

        return $this->remoteIndexTemplate;
    }

    /**
     * @return IndexerConfigurationInterface
     * @throws NoSuchEntityException
     */
    public function getIndexerConfiguration(): IndexerConfigurationInterface
    {
        if ($this->indexerConfiguration === null) {
            /** @var IndexerConfigurationInterface indexerConfiguration */
            $this->indexerConfiguration = $this->indexerConfigurationRepository->getByIndexerId($this->getName());
        }

        if ($this->indexerConfiguration === null) {
            $this->indexerConfiguration = $this->indexerConfigurationFactory->create()
                ->setEnabled(false);
        }

        return $this->indexerConfiguration;
    }

    /**
     * @return string
     * @throws NoSuchEntityException
     */
    protected function getName()
    {
        if ($this->name === null) {
            $this->name = $this->indexerResolver->resolve($this->getClass());
        }

        return $this->name;
    }

    /**
     * @return string
     */
    abstract protected function getClass(): string;

    /**
     * @return Connector\ConnectorInterface
     * @throws NoSuchEntityException
     */
    public function getConnector(): Connector\ConnectorInterface
    {
        return $this
            ->getIndexerConfiguration()
            ->getConnectorConfiguration()
            ->getConnector()
            ->setMetadata($this->prepareConnectorMetadata($this->storeManager->getStore()));
    }

    /**
     * @param StoreInterface $store
     * @return array
     */
    public function prepareConnectorMetadata(StoreInterface $store): DataObject
    {
        return $this->dataObjectFactory->create();
    }

    /**
     * @return LoggerInterface
     */
    protected function getLogger()
    {
        return $this->logger;
    }

    /**
     * @param $entities
     * @param $collection
     * @return mixed
     */
    abstract protected function write($entities, $collection);
}
